@@ -60,23 +60,23 @@ namespace snmalloc
   {
     /**
      * Flag to protect the bump allocator
-     **/
+     */
     std::atomic_flag lock = ATOMIC_FLAG_INIT;
 
     /**
      * Pointer to block being bump allocated
-     **/
+     */
     void* bump = nullptr;
 
     /**
      * Space remaining in this block being bump allocated
-     **/
+     */
     size_t remaining = 0;
 
     /**
      * Simple flag for checking if another instance of lazy-decommit is
      * running
-     **/
+     */
     std::atomic_flag lazy_decommit_guard = {};
 
   public:
@@ -87,7 +87,7 @@ namespace snmalloc
 
     /**
      * Make a new memory provide for this PAL.
-     **/
+     */
     static MemoryProviderStateMixin<PAL>* make() noexcept
     {
       // Temporary stack-based storage to start the allocator in.
@@ -125,6 +125,12 @@ namespace snmalloc
     {
       // Reserve the smallest large_class which is SUPERSLAB_SIZE
       void* r = reserve<false>(0);
+
+      if (r == nullptr)
+        Pal::error(
+          "Unrecoverable internal error: \
+          failed to allocator internal data structure.");
+
       PAL::template notify_using<NoZero>(r, OS_PAGE_SIZE);
 
       bump = r;
@@ -197,7 +203,7 @@ namespace snmalloc
 
     /***
      * Method for callback object to perform lazy decommit.
-     **/
+     */
     static void process(PalNotificationObject* p)
     {
       // Unsafe downcast here. Don't want vtable and RTTI.
@@ -274,6 +280,9 @@ namespace snmalloc
         size_t request = bits::max(size * 4, SUPERSLAB_SIZE * 8);
         void* p = PAL::template reserve<false>(request);
 
+        if (p == nullptr)
+          return nullptr;
+
         address_t p0 = address_cast(p);
         address_t start = bits::align_up(p0, align);
         address_t p1 = p0 + request;
@@ -354,6 +363,8 @@ namespace snmalloc
       if (p == nullptr)
       {
         p = memory_provider.template reserve<false>(large_class);
+        if (p == nullptr)
+          return nullptr;
         memory_provider.template notify_using<zero_mem>(p, size);
       }
       else
